/*****************************************************************************
 *
 * cdmknods.c 
 *
 * builds device, diag, and line nodes for Digi International ELS ports
 *	 Usage: -auto                        configure ports
 *	 Usage: -addunit{ name type IPaddr}  configure additional unit
 *		Additional options:
 *		 -dryrun    check els.conf file
 *		 -silent  disables non-error messages
 *               -help      displays usage info
 *		 -nhandle   suffix for normal ports e.g. c  ttyc and lpc dev
 *               -dhandle   suffix for dialout ports
 *               -nodialout don't create any dialout nodes
 *               -noterm    don't create any normal terminal nodes
 *               -inittab   create entries to append to inittab
 *               -ibaud     baud rate for inittab if not 9600
 *               -launch    start the cdetherd daemons for all units
 *            
 *****************************************************************************
 *
 * Copyright 2000 Digi International (www.digi.com)
 *    Bob Rubendunst <Robert_Rubendunst@digi.com>
 *    Jeff Randall   <Jeff_Randall@digi.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/sysmacros.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pwd.h>
#include "ctio.h"
#include "sts_unittab.h"

#ifndef LOCAL
# define LOCAL extern
#endif

#ifndef NAMELEN
# define NAMELEN 16
#endif

#include "cdutils.h"

/* 
 * we can define a total of 7 units per major on up to 9 majors
 * for a total of 63 units. Each unit can have up to 32 ports,
 * for a maximum total of 2016 ports.
 */
#define MAX_ELSDEVS	9*7
#define ELS_GLOBAL_MINOR 255

#ifndef ELS_CONF
# define ELS_CONF	"/etc/els.conf"
#endif

#ifndef DEV_PATH
# define DEV_PATH	"/dev/dg/els"
#endif

#if !defined(MAX)
# define MAX(x,y)	(x > y ? x : y)
#endif

#if !defined(MIN)
# define MIN(x,y)	(x < y ? x : y)
#endif


/* global vars */
int max_pun_scanned	= -1;	/* highest els unit # defined in els.conf */
char *nsuffix		= "N";	/* for tty or lp (defaults to ttyN and lpNc) */
char *dsuffix		= "n";	/* for dial out nodes (defaults to cun) */
int ELS_fd;
int cdp_fd;
int bus			= 0;	/* SCSI bus # (0 or 1) */
int target		= -1;	/* SCSI target ID */
int nodialout		= 0;
int noterm		= 0;

int cdp_majorno	= -1;		/* major device number */
int major_set	= FALSE;	/* flag for major specified */
int verbose	= TRUE;		/* default to being noisy of course */
int launch	= FALSE;	/* start cdetherds */
int replace	= TRUE; 	/* remove devices and replace if they exist */
int uninstall	= FALSE;	/* delete all of our nodes for pkg uninstall */
int pci_only	= FALSE;	/* make only PCI nodes */
int nodes_made	= 0;		/* count of nodes created */
int automatic	= FALSE;	/* automatic mode */
int build_gettys = FALSE;
char *gettys_file = "gettys.els";
FILE *getty_fp	= NULL;
char *getty_baud = "9600";	/* default inittab baud string */
int dryrun	= FALSE;	/* just checking, don't make anything */
int addunit	= FALSE;	/* adding a new unit on the fly */
int cd_unit_cnt	= 0;		/* number of units found in auto mode */
struct stat statbuf;		/* to check and see if node already exists */
struct ctunitinfo *uinfo;	/* unittab record */
char units[MAX_ELSDEVS+1][17];	/* pointers for all possible units */
int auto_pun[MAX_ELSDEVS+1];	/* device # for each ELS unit */
#define IPADDRSZ 17
char auto_ip[MAX_ELSDEVS+1][IPADDRSZ];	/* IP address for each ELS unit */
int configured[MAX_ELSDEVS];	/* for testing existence in -addunit */
gid_t tty_gid	= 0;
gid_t uucp_gid	= 0;

gid_t daemon_gid = 0;
int first_major	= -1;
char rad32[]="0123456789abcdefghijklmnopqrstuv"; /* for last character of
						  * port name */

/* prototypes */
int main(int argc, char **argv );
void do_args( int argc, char **argv );
void usage(int usage_code );
int make_tty_node(int majorno, int minorno, char *name,
		  mode_t mode, uid_t own, uid_t grp);
int make_many_nodes(int id, struct ctunitinfo *uinfo);
int remove_many_nodes(int id, struct ctunitinfo *uinfo);
int make_auto_nodes();
int teardown();
int add_unit();
void read_els_conf(int index);
int make_els_unit_node(int modmajor, int  modminor, char *name);
char *get_full_name(char *arg);
int make_diag_node(char *node, int majorno, int minorno);
int remove_diag_node(char *node);
void inittab_line(char *name);
int start_cdetherds(int count);
int make_ourdir(char *name);
int remove_ourdir(char *name);





/* 
 *	main
 */
int main(int argc, char **argv) {
	
	struct passwd *p;
	char uucp[]="uucp";
	char daemon[]="daemon";
	char tty[]="tty";
	int rc=0;
	int daemons_started = 0;

	/* must run as root */
	if (geteuid() != 0) {
		fprintf(stderr, "You must be root to use cdmknods\a\n");
		usage(1);
	}

	/* sanity check the arguments */
	if (argc < 2) 
		usage(1);

	/* process argument list */
	do_args(argc, argv);

	if ( addunit +  uninstall +automatic > 1) {
		fprintf(stderr, "Can only specify one of: -auto, "
			"-addunit or -uninstall.\n");
		usage(1);
	}

	if (build_gettys) {
		if (strcasecmp(gettys_file,"/etc/inittab") == 0) {
			fprintf(stderr, "Overwriting /etc/inittab not "
				"supported.\n");
			exit(1);
		}
		if ((getty_fp = fopen(gettys_file,"w+")) == NULL) {
			fprintf(stderr, "Error opening getty file %s:%s\n",
				gettys_file, strerror(errno));
			build_gettys = FALSE;
		}
		if (!build_gettys)
			fprintf(stderr, "Output of inittab gettys "
				"cancelled.\a\n");
	}

	p = getpwnam(uucp);
	if (p)
		uucp_gid = p->pw_gid;
	p = getpwnam(tty);
	if (p)
		tty_gid = p->pw_gid;
	p = getpwnam(daemon);
	if (p)
		daemon_gid = p->pw_gid;

	umask(0);

	if (uninstall) {
		rc |= teardown();
		rc |= remove_diag_node(DEV_DIAG);
		rc |= remove_ourdir(DEV_PATH);
	}
	if (automatic) {
		rc |= make_ourdir(DEV_PATH);
		rc |= make_auto_nodes();
		rc |= make_diag_node(DEV_DIAG, first_major, ELS_GLOBAL_MINOR);
	}
	if (addunit) {
		rc |=  make_ourdir(DEV_PATH);
		rc |= add_unit();
		if (stat(DEV_DIAG, &statbuf) != 0)
			rc |= make_diag_node(DEV_DIAG, first_major,
					     ELS_GLOBAL_MINOR);
	}
	if (!rc && launch)
		daemons_started  = start_cdetherds(cd_unit_cnt);

	/* tell 'em what we did */
	if (verbose) {
		if (nodes_made > 3) {
			fprintf(stdout, "  Done, made %u nodes.\n",
				nodes_made);
			fprintf(stdout, "        started %d cdetherd "
				"daemons.\n", daemons_started);
		} 
		else
			fprintf(stdout, "  Done\n");
	}
	if (getty_fp)
		fclose(getty_fp);
	exit(0);
	/* "Computer: End Program." -Lt. Barkley, ST:TNG Ship in a bottle */
}


/*
 * make_tty_node()
 * create a tty node. If this is not a dry run, return non-zero if the
 * node could not be created.
 */
int make_tty_node(int majorno, int minorno, char *name,
		  mode_t mode, uid_t own, uid_t grp) {

	if (stat(name,&statbuf) == 0) {
		if (replace != TRUE) {
			fprintf(stderr,
				"Node '%s' already exists. You must remove "
				"it first to change it.\n",name);
			return((dryrun) ? 0 : 1);
		}
		else {
			if (!dryrun && (unlink(name) == -1)) {
				fprintf(stderr,
					"Node '%s' exists and could not "
					"be removed: ", name);
				perror(NULL);
				return((dryrun) ? 0 : 1);
			}
		}
	}
	/* actually make the node */
	if (dryrun) /* verbose */
		fprintf(stdout, "Making node %s, major=%u minor=0x%x\n", 
			name, majorno, minorno);
  
	if (!dryrun && mknod(name, (S_IFCHR | mode), 
			     makedev(majorno, minorno)) < 0) {
		fprintf(stderr,"Error making node %s\n.", name);
		perror("cdmknods");
		return(1);
	}
	
	/* change the ownership */
	if (!dryrun && (own | grp)) {
		if (chown(name, own , grp) != 0) {
			fprintf(stderr, "Error changing ownership of "
				"node %s\n", name);
			perror("cdmknods");
			return(1);
		}
	}
	nodes_made++;
	return(0);
}

/*
 *   Make the unit node and the line nodes for a device id.
 *   Because we can support more ELS ports than minor devices,
 *   we need some scheme to handle all the ports. Here is how
 *   it works:
 *
 *   Unit   /proc/devices entry    ELS unit name    	ELS port names 
 *   0      ttyN0 / cun0           /dev/dg/els/els00    /dev/ttyN000 - ttyN00v
 *   1      ttyN0 / cun0           /dev/dg/els/els01    /dev/ttyN010 - ttyN01v
 *   [...]
 *   6      ttyN0 / cun0           /dev/dg/els/els06    /dev/ttyN060 - ttyN06v
 *   7      ttyN1 / cun1           /dev/dg/els/els10    /dev/ttyN100 - ttyN10v
 *   8      ttyN1 / cun1           /dev/dg/els/els11    /dev/ttyN100 - ttyN11v
 *   [...]
 *   55     ttyN8 / cun8           /dev/dg/els/els76    /dev/ttyN760 - ttyN76v
 *
 */

/*
 * find_proc_major reads the /proc/devices file to find the
 * major device number assigned to the module at insmod time.
 */
int find_proc_major(char *file, char *key) {

#define LNBUFSZ 200
	char linbuf[LNBUFSZ+1];
	char *srch;
	FILE *fd;
	int found_major = -1;

	fd = fopen(file, "r");
	if (fd == NULL) 
		return(found_major);	

	while (!feof(fd) && fgets(linbuf, LNBUFSZ - 1, fd) != NULL) {
		srch = strchr(linbuf, '\n');
		if (srch != NULL)	
			*srch = 0;
		else
			continue;
		if ((srch = strstr(linbuf, key)) == NULL)
			continue;
		found_major = atoi(linbuf);
		break;
	}
	fclose(fd);
	return(found_major);
}


/* 
 * make_auto_nodes()
 *
 * Read the els.conf file, building arrays of unit names and IPs.
 * Build the nodes for all each unit in the arrays.
 */
int make_auto_nodes()
{
	int i, rc;
	struct ctunitinfo *uinfo;

	read_els_conf(0);
	for (i=0; i < cd_unit_cnt; i++) {
		if (NULL == (uinfo = ctm_get_unitinfo(units[i]))) {
			fprintf(stderr, "unknown unit type %s\n",units[i]);
			return 1;
		}
		if ((rc=make_many_nodes(auto_pun[i], uinfo )))
			break;
	}
	return(rc);
}


/* 
 * teardown()
 *
 * Read the els.conf file, building arrays of unit names and IPs.
 * Remove the nodes for all each unit in the arrays.
 */
int teardown() {

	int i, rc;
	struct ctunitinfo *uinfo;

	read_els_conf(0);
	for (i=0; i < cd_unit_cnt; i++) {
		if (NULL == (uinfo = ctm_get_unitinfo(units[i]))) {
			fprintf(stderr, "unknown unit type %s\n",units[i]);
			return(1);
		}
		if ((rc=remove_many_nodes(auto_pun[i], uinfo )))
			break;
	}
	return(rc);
}


/*
 * Add_unit()
 *
 * Read in the els.conf file, but starting at index 1 to leave
 * the unit name and IP address set in do_args() intact. 
 * Check to make sure the unit is not already defined. If not,
 * append its entries to the els.conf file, then build the unit nodes.
 */
int add_unit() {

	int rc;
	struct ctunitinfo *uinfo;
	FILE *fd;

	/* set array of configured units */
	read_els_conf(1);
	printf("adding unit %s\n", units[0]);
	if (NULL == (uinfo = ctm_get_unitinfo(units[0]))) {
		fprintf(stderr, "unknown unit type-%s\n",units[0]);
		return(1);
	}
	if (auto_pun[0] > MAX_ELSDEVS) {
		fprintf(stderr, "els%x exceeds maximum unit els%x\n",
			auto_pun[0], MAX_ELSDEVS);
		return(1);
	}
	if (configured[auto_pun[0]]) {
		fprintf(stderr, "els%2.2x already defined els.conf line %d\n",
			auto_pun[0], configured[auto_pun[0]]);
		return(1);
	}
	if ( 0 >= inet_addr(auto_ip[0])) {
		fprintf(stderr,"bad IP address %s\n",auto_ip[0]);
		return(1);
	}
	if ( NULL == (fd = fopen(ELS_CONF, "a+")) ||
	     ( 0 > fprintf(fd, "els%x %s %s\n", auto_pun[0],
			   units[0], auto_ip[0]))) {
		fprintf(stderr,"-addunit failed on update els.conf: %s\n",
			strerror(errno));
		return(1);
	}
	rc=make_many_nodes(auto_pun[0], uinfo );
	/* if addunit worked, start up the daemon, too. */
	if (!rc) {
		cd_unit_cnt = 1;
		launch = TRUE;
	}
	return(rc);
}


/* 
 * make all the serial and printer ports for a unit.
 * We need the unit number 0-63 and a ptr to the unit info struct.
 */
int make_many_nodes(int id, struct ctunitinfo *uinfo ) 
{
	static char temp[80];
	int mod = id / 7;
	int unit = id % 7;
	int type;
	int modmajor = -1;
	int cunmajor = -1;
	int rc=0;
	int line;
	
	if (id >= MAX_ELSDEVS) {
		fprintf(stderr, "unit number must be less than %d\n",
			MAX_ELSDEVS);
		return(1);
	}
	/*
	 * Use the unit name registered by the module to get the
	 * MAJOR numbers for this device
	 */
	sprintf(temp, "ttyN%d", mod);
	modmajor = find_proc_major("/proc/devices", temp);
	if (first_major < 0 && modmajor > 0)
		first_major = modmajor;	/* save for diag node later */
	sprintf(temp, "cun%d", mod);
	cunmajor = find_proc_major("/proc/devices", temp);
	if (modmajor < 0 || cunmajor < 0) {
		fprintf(stderr, "can't find installed els module device "
			"ttyN%d or cun%d.\n",mod, mod );
		return(1);
	}
	/* the ELS unit name */
	sprintf(temp, "%s/els%2.2x", DEV_PATH, id);	
	if (make_els_unit_node(modmajor, 224 + unit, temp)) {
		fprintf(stderr, "can't create %s device node.\n", temp);
		return(1);
	}
	for (line=0; line < uinfo->num_lines; line++) {
		type = ctf_port_type(uinfo, line);
		if (type == PT_NODEV)
			continue;
		if (PT_IS_PARALLEL(type)) {
			sprintf(temp,"%s/lp%s%2.2x8", DEV_PATH, nsuffix, id);
			if ((rc=make_tty_node(modmajor, unit*32+line,
					      temp, 0660, 0, daemon_gid)))
				break;
		} else {
			/* Make Normal nodes both in /dev and DEV_PATH */
			sprintf(temp,"%s/tty%s%2.2x%c", "/dev", nsuffix, id,
				rad32[line]);
			if ((rc=make_tty_node(modmajor, unit*32+line,
					      temp, 0664, 0, tty_gid)))
				break;
			sprintf(temp,"%s/tty%s%2.2x%c", DEV_PATH, nsuffix, id,
				rad32[line]);
			if ((rc=make_tty_node(modmajor, unit*32+line,
					      temp, 0664, 0, tty_gid)))
				break;

			/* just make callout nodes in DEV_PATH */
			sprintf(temp,"%s%x%d%c", nsuffix, mod, unit,
				rad32[line]);
			inittab_line(temp);
			sprintf(temp,"%s/cu%s%2.2x%c", DEV_PATH, dsuffix,
				id, rad32[line]);
			if ((rc=make_tty_node(cunmajor, unit*32+line,
					      temp, 0600, 0, 0)))
				break;
		}
	}
	return(rc);
}


/* 
 * remove all the serial and printer ports for a unit.
 * We need the unit number 0-63 and a ptr to the unit info struct.
 */
int remove_many_nodes(int id, struct ctunitinfo *uinfo ) {

	static char temp[80];
	int type;
	int rc=0;
	int line;
	
	if (id >= MAX_ELSDEVS) {
		fprintf(stderr, "unit number must be less than %d\n",
			MAX_ELSDEVS);
		return(1);
	}

	/* the ELS unit name */
	sprintf(temp, "%s/els%2.2x", DEV_PATH, id);	
	if (!dryrun && unlink(temp)) {
		fprintf(stderr, "can't remove %s device node,%s.\n",
			temp, strerror(errno));
	}
	for (line=0; line < uinfo->num_lines; line++) {
		type = ctf_port_type(uinfo, line);
		if (type == PT_NODEV)
			continue;
		if (PT_IS_PARALLEL(type)) {
			sprintf(temp,"%s/lp%s%2.2x8", DEV_PATH, nsuffix, id);
			if (!dryrun && unlink(temp))
				fprintf(stderr, "can't remove %s device "
					"node,%s.\n", temp, strerror(errno));
		} else {
			/* remove the normal nodes from /dev and DEV_PATH */
			sprintf(temp,"%s/tty%s%2.2x%c", "/dev", nsuffix,
				id, rad32[line]);
			if (!dryrun && unlink(temp))
				fprintf(stderr, "can't remove %s device node: "
					"%s.\n", temp, strerror(errno));
			sprintf(temp,"%s/tty%s%2.2x%c", DEV_PATH, nsuffix,
				id, rad32[line]);
			if (!dryrun && unlink(temp))
				fprintf(stderr, "can't remove %s device node: "
					"%s.\n", temp, strerror(errno));

			/* remove callout nodes just from DEV_PATH */
			sprintf(temp,"%s/cu%s%2.2x%c", DEV_PATH, dsuffix, id,
				rad32[line]);
			if (!dryrun && unlink(temp))
				fprintf(stderr, "can't remove %s device node: "
					"%s.\n", temp, strerror(errno));
		}
	}
	return(rc);
}


/*
 * read_els_conf()
 *
 * read all the ELS units from the config file.
 * Load them into the array starting at index.
 * Index is set to 1 in -addunit to leave array's zero contents alone.
 * Keep track of the maximum unit # configured.
 * (They do not have to be defined in order.)
 * Store the values in an array.
 *
 */
void read_els_conf(int index) {

#define MAX_CONF_ENTRY 200
	char conf_entry[MAX_CONF_ENTRY+1]; /* space to parse config file in */
	int pun;
	int count=0;
	FILE *fd;			/* file handle for /etc/els.conf */
	char ip[33];
	char elstype[18];
	char *temp;

	int line = 0;
	fd = fopen(ELS_CONF, "r");
	if (fd == NULL) {
		return;	/* couldn't open - assume no etherlite devices */
	}
	
	if (verbose)
		fprintf(stdout, "Scanning %s for Digi International ELS "
			"devices...\n", ELS_CONF);
     
	while ((count+index) < MAX_ELSDEVS && !feof(fd) && 
	       (fgets(conf_entry, MAX_CONF_ENTRY, fd) != NULL)) {
		line += 1;
		temp = strchr(conf_entry, '\n');
		if (temp != NULL)	
			*temp = 0;
		else 
			break;

		if (strncasecmp(conf_entry, "els", 3)) {
			if (verbose) {
				if (conf_entry[0] != '#')
					fprintf(stdout,
						"skipping %s\n", conf_entry);
				continue;
			}
		}

		if (sscanf(conf_entry, "els%x %17s %32s", &pun,
			   &elstype[0], &ip[0]) !=3) {
			if (verbose)
				fprintf(stderr, "ignoring line with bad "
					"format: %s\n", conf_entry);
			continue;
		}
		
		auto_pun[count+index] = pun;
		if (pun > max_pun_scanned)
			max_pun_scanned = pun;
		configured[pun] = line;	/* for -addunit check & error msg */
		temp = get_full_name(elstype);
		strncpy((char *)units[count+index], temp, 17);
		strncpy(auto_ip[count+index], ip, IPADDRSZ);
		if (verbose) {
			fprintf(stderr, "Etherlite(R) Server %2.2x: %s "
				"at %s\n", pun, units[count+index], ip);
			fflush(stdout);
		}
		
		count++;
	}
	fclose(fd);
	
	cd_unit_cnt = count;
}


int make_diag_node(char *node, int majorno, int minorno) {

	int i;
	int rc = 0;

	if (majorno == -1)
		return 1;

	i = stat(node,&statbuf);
	if (i == 0) {
		if (replace == TRUE) 
			if (!dryrun && unlink(node) == -1) {
				fprintf(stderr,"\nNode '%s' exists and "
					"could not be removed: ", node);
				perror(NULL);
				exit(1);
			}
	}
	
	if ((i != 0) || (replace == TRUE)) {
		if (!dryrun && (mknod(node, S_IFCHR | 0644, 
				      makedev(majorno, minorno)) != 0)) {
			fprintf(stderr,"\nError making node %s\n", node);
			exit(1);
		}
		else if (dryrun)		/* verbose */
			fprintf(stderr,"Created diag node %s\n", node);
	}
	return(rc);
}


int remove_diag_node(char *node) {

	int i;
	int rc = 0;

	i = stat(node,&statbuf);
	if (i == 0) {
		if (!dryrun)
			rc=unlink(node);
	}
	return(rc);
}

/* make sure our custom node directory exists. */
int make_ourdir(char *name) {
char *partial= NULL;
char *slash;

	if (stat(name,&statbuf) != 0) {
		partial = strdup(name);
		if (partial && ((slash=strrchr(partial, '/')) != NULL)) {
			*slash = 0;
			if (stat(partial,&statbuf) != 0) {
				if (!dryrun && mkdir(partial, 0755) == -1) {
					fprintf(stderr,	"could not create '%s': ", name);
					perror(NULL);
					free(partial);
					return(1);
				}
			}
				
		}
		free(partial);
		if (!dryrun && mkdir(name, 0755) == -1) {
			fprintf(stderr,	"could not create '%s': ", name);
			perror(NULL);
			return(1);
		}
	}
return 0;
}

/* make sure our custom node directory vanishes. */
int remove_ourdir(char *name) {
	if (stat(name,&statbuf) == 0) {
		if (!dryrun && unlink(name) == -1) {
			fprintf(stderr,	"could not remove '%s': ", name);
			perror(NULL);
			return(1);
		}
	}
return 0;
}

/*
 *
 * Make the etherlite unit nodes
 *
 */
int make_els_unit_node(int modmajor, int  modminor, char *name) {

	/* see if node already exists - always replace */
	if (stat(name,&statbuf) == 0) {
		if (!dryrun && unlink(name) == -1) {
			fprintf(stderr,	"\nNode '%s' exists and could not "
				"be removed: ", name);
			perror(NULL);
			return(1);
		}
	}
  
	if (!dryrun && mknod(name, (S_IFCHR | 0600), 
			     makedev(modmajor, modminor)) < 0) {
		fprintf(stderr,"\nError making node %s\n.", name);
		perror("cdmknods");
		return(1);
	}
	else if (dryrun)		/* verbose */
		fprintf(stderr, "\nCreated unit node %s\n", name);
	return(0);
}


/*
 * do_args()
 *
 * parse the mess of a command line out and make sure everything is sane.
 *
 */
void do_args (int argc, char **argv) {
 	
	int ar=0;
	
	/* scan arg list */
	for (ar = 1; ar < argc; ar++) {
		
#if 0
		/* set mode */
		if (strcasecmp(argv[ar],"-mode") == 0) {
			/* allow multiple modes */
			ar++;
			if ((strcasecmp(argv[ar],"local") == 0) || 
			    (strcasecmp(argv[ar],"hardwired") == 0)) {
				mode = DIRECT;
				mode_set = TRUE;
				continue;
			}
			/* dial-out mode */
			if (strcasecmp(argv[ar],"dialout") == 0) {
				mode = DIALOUT;
				mode_set = TRUE;
				continue;
			}
			
			/* catch bad mode spec */
			fprintf(stderr,
				"\nOption -mode expects param: local "
				"or dialout.\n");
			usage(1);
		}
#endif /* 0 */
		
		/* set normal handle, i.e. N=ttyN & lpNc, Z=ttyZ and lpZ.... */
		if (strcasecmp(argv[ar],"-nhandle") == 0) {
			ar++;
			nsuffix = argv[ar];
			if (strlen(nsuffix) >= 4) {
				fprintf(stderr,
					"\nDevname '%s%s*' is too long.\n"
					"nhandle should be no "
					"longer than 4 characters.\n",
					DEV_PATH"/tty", nsuffix);
				exit(1);
			}
			continue;
		}
		if (strcasecmp(argv[ar],"-dhandle") == 0) {
			ar++;
			dsuffix = argv[ar];
			if (strlen(dsuffix) >= 4) {
				fprintf(stderr,
					"\nDevname '%s%s*' is too long.\n"
					"dnandle should be no "
					"longer than 4 characters.\n",
					DEV_PATH"/tty", dsuffix);
				exit(1);
			}
			continue;
		}
		if (strcasecmp(argv[ar],"-inittab") == 0) {
			ar++;
			build_gettys = TRUE;
			gettys_file = argv[ar];
			if (ar >= argc)
				break;
			continue;
		}
		if (strcasecmp(argv[ar],"-ibaud") == 0) {
			ar++;
			getty_baud = argv[ar];
			continue;
		}
		if (strcasecmp(argv[ar],"-silent") == 0) {
			verbose = FALSE;
			continue;
		}
		if (strcasecmp(argv[ar],"-launch") == 0) {
			launch = TRUE;
			continue;
		}
		if (strcasecmp(argv[ar],"-dryrun") == 0) {
			dryrun = TRUE;
			continue;
		}
		if (strcasecmp(argv[ar],"-uninstall") == 0) {
			uninstall = TRUE;
			continue;
		}
		if (strcasecmp(argv[ar],"-addunit") == 0) {
			ar++;
			if ((sscanf(argv[ar++],"els%x", &auto_pun[0]) ==1) &&
			    (sscanf(argv[ar++],"%17s", &units[1][0])==1) &&
			    (sscanf(argv[ar],"%32s", &auto_ip[0][0]) == 1));
			else {
				fprintf(stderr, "-addunit requires unit name, "
					"type, IP address.\n");
				exit(1);
			}
			/* normalize the unit name and put in ele 0 */
			strncpy((char *)units[0], get_full_name(units[1]), 17);
			
			addunit = TRUE;
			continue;
		}
		if (strcasecmp(argv[ar],"-auto") == 0) {
			if (addunit || uninstall) {
				fprintf(stderr, "Cannot specify -auto with "
					"-addunit or -uninstall.\n");
				usage(1);
			}
			else {
				automatic = TRUE;
				continue;
			}
		}
		if (strcasecmp(argv[ar],"-check") == 0) {
			dryrun = TRUE;
			continue;
		}
		if (strcasecmp(argv[ar],"-noerase") == 0) {
			replace = FALSE;
			continue;
		}
		
		if (strcasecmp(argv[ar],"-help") == 0) {
			usage(0);
		}
		
		/* catch defaults */
		fprintf(stderr, "\nUnrecognized option '%s'.\n",argv[ar]);
		usage(1);
	}
	
	/*
	 * make sure we have reasonable info
	 */
	
}


/*
 * usage();
 *
 */
void usage(int usage_code) {

	int i;
	static char *use[] = {
"",
"Usage: cdmknods -auto"
"",
"     Supported products: EL-2, EL-8+, EL-16, EL-32",
"",
"",
"     Additional options:",
"     -dhandle [suffix] suffix for tty and lp port nodes",
"     -nhandle [suffix] suffix for dialout port nodes",
"     -nodialout        don't create dialout nodes",
"     -noterm           don't create terminal nodes",
"     -silent           disables non-error messages",
"     -noerase          don't remove and replace nodes that already exist",
"     -launch           launch cdetherd daemons after building nodes",
"     -inittab file     create file with getty entries",
"     -ibaud baud       specify baud rate other than 9600 for above",
"     -dryrun           don't build nodes or launch cdetherd(s)",
"     -help             displays this usage info",
"     cdmknods -addunit elsxx el-type ip-addr",
"",
NULL };

	for (i = 0; use[i]; i++)
		fprintf(stderr,"%s\n",use[i]);
	
	exit(usage_code);
}




/*
 * get_full_name()
 *
 * turn partial names into full names that ELS_unittab recognizes
 *
 * For cdmknods, only the number and types of ports matter - fast/slow does 
 * not as the driver will determine that at usage time
 *
 */
char *get_full_name(char *arg) {

	int i;
	static char part[NAMELEN+1];

	bzero(part, NAMELEN+1);
	strncpy(part, arg, NAMELEN);
	for (i=0; i < NAMELEN; i++) {
		part[i] = toupper(part[i]);
	}
  
	if (strcasecmp(part, "EL2") == 0)		return("EL-2");
	if (strcasecmp(part, "EL2.1") == 0)		return("EL-2");
	if (strcasecmp(part, "EL2.2") == 0)		return("EL-2");
	if (strcasecmp(part, "EL-2-EIA485") == 0)	return("EL-2");
	if (strcasecmp(part, "EL8+") == 0)		return("EL-8+");
	if (strcasecmp(part, "EL8P") == 0)		return("EL-8+");
	if (strcasecmp(part, "EL-8P") == 0)		return("EL-8+");
	if (strcasecmp(part, "EL16") == 0)		return("EL-16");
	if (strcasecmp(part, "EL160") == 0)		return("EL-160");
	if (strcasecmp(part, "EL162") == 0)		return("EL-162");
	if (strcasecmp(part, "EL32") == 0)		return("EL-32");
	if (strcasecmp(part, "\"EL2\"") == 0)		return("EL-2");
	if (strcasecmp(part, "\"EL2.1\"") == 0)		return("EL-2");
	if (strcasecmp(part, "\"EL2.2\"") == 0)		return("EL-2");
	if (strcasecmp(part, "\"EL-2-EIA485\"") == 0)	return("EL-2");
	if (strcasecmp(part, "\"EL8+\"") == 0)		return("EL-8+");
	if (strcasecmp(part, "\"EL8P\"") == 0)		return("EL-8+");
	if (strcasecmp(part, "\"EL-8P\"") == 0)		return("EL-8+");
	if (strcasecmp(part, "\"EL16\"") == 0)		return("EL-16");
	if (strcasecmp(part, "\"EL160\"") == 0)		return("EL-160");
	if (strcasecmp(part, "\"EL162\"") == 0)		return("EL-162");
	if (strcasecmp(part, "\"EL32\"") == 0)		return("EL-32");
	if (strcasecmp(part, "\"EL-8+\"") == 0)		return("EL-8+");
	if (strcasecmp(part, "\"EL-16\"") == 0)		return("EL-16");
	if (strcasecmp(part, "\"EL-32\"") == 0)		return("EL-32");
  
	return(part);  /* we don't have an abbreviation for it, pass it back */
}

/*
 * inittab_line()
 *
 * outputs an inittab style line if we are making an inittab append file.
 * Otherwise it just returns.
 */
void inittab_line(char *name) {

	if (build_gettys)
		fprintf(getty_fp, "%s:2345:respawn:/sbin/getty tty%s %s\n",
			name, name, getty_baud);
}

/*
 * start_cdetherds()
 *
 * start the cdetherd daemons for the units picked up by -auto,
 * unless the dryrun flag is on. */
int start_cdetherds(int count) {

	int i, mod , unit;
	int started = 0;
	char cmd[80];

	for (i=0; i < count ; i++) {
		mod = auto_pun[i] / 7;
		unit = auto_pun[i] % 7;
		sprintf(cmd,"cdetherd els%2.2x %s", auto_pun[i], auto_ip[i]);
		if (verbose || dryrun)
			fprintf(stderr, "starting %s\n",cmd);
		if (!dryrun && (system(cmd) == 0))
			started += 1;
	}
	return(started);
}

/* end of cdmknods.c */
