/*
 * SCSI engine for processing FAS commands
 *
 * Copyright (C) 2000,2001 Kees Cook
 * cook@cpoint.net, http://outflux.net/
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * http://www.gnu.org/copyleft/gpl.html
 *
 */
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <scsi/sg.h>
#include <sys/ioctl.h>  /* ioctl */
#include <ctype.h>	/* isprint */

#define __USE_XOPEN
#define __USE_GNU
#include <stdlib.h>

#include "scsi_iface.h"
#include "scsi_tools.h"

#include "sts.h"
void fas_packet_ret(sts_state * sts, unsigned char * cmd);
void fas_error(FILE * fd, unsigned int status);

/* this is to have the UNIT_POLL response to have error bits cleared */
#define UNIT_POLL_HACK

/* helper macros for 16bit Least significant bit first */
#define GRAB_16BIT(ptr, start) ((ptr)[(start)] + ((ptr)[(start)+1]<<8))
#define STORE_16BIT(ptr,start,val) \
			(ptr)[(start)]=((val) & 0xff); \
			(ptr)[(start)+1]=(((val) & 0xff00) >> 8);

/* helper macros for bit-title manipulations: bit# 0, 1, ... 7 */
#define IS_SET(byte, bit)	(((byte) & (1 << (bit))) == (1 << (bit)))
#define SET(byte, bit)		((byte) | (1 << (bit)))
#define UNSET(byte, bit)	((byte) & (~(1 << (bit))))



typedef struct fas_opcode_info_t {
	char * t;		/* text name */
	unsigned char req_family;/* family of FAS commands it belongs to */
	unsigned char rsp_family;/* family of FAS commands it belongs to */
} fas_opcode_info;

fas_opcode_info fas_opcodes[] = {
        { "FAS_GLOBAL", FAS_GLOBAL, FAS_GLOBAL },
        { "FAS_ENABLE", FAS_ENABLE, FAS_ENABLE },
        { "FAS_DISABLE", FAS_ENABLE, FAS_ENABLE },
        { "FAS_SEND", FAS_SEND, FAS_ENABLE },
        { "FAS_RECV", FAS_ENABLE, FAS_RECV },
        { "FAS_IN_TIMERS", FAS_IN_TIMERS, FAS_ENABLE },
        { "FAS_OUTPUT_CTL", FAS_OUTPUT_CTL, FAS_ENABLE },
        { "FAS_INPUT_CTL", FAS_OUTPUT_CTL, FAS_ENABLE },
        { "FAS_SET_MODEM", FAS_SET_MODEM, FAS_ENABLE },
        { "FAS_STAT_CHG", FAS_STAT_CHG, FAS_STAT_CHG },
        { "FAS_SEND_BRK", FAS_ENABLE, FAS_ENABLE },
        { "FAS_SET_PARAMS", FAS_SET_PARAMS, FAS_ENABLE },
        { "FAS_FLOW_CTL", FAS_FLOW_CTL, FAS_ENABLE },
        { "FAS_RESERVE", FAS_ENABLE, FAS_ENABLE },
        { "FAS_RELEASE", FAS_ENABLE, FAS_ENABLE },
	{ "FAS_RESET", FAS_ENABLE, FAS_ENABLE },	/* 15 */
	{ "FAS_UNIT_STATS", FAS_UNIT_STATS, FAS_RECV },	/* 16 */
	{ "FAS_WAIT_SEND", FAS_ENABLE, FAS_ENABLE },	/* 17 */
	{ "FAS_LINE_STATE", FAS_UNIT_STATS, FAS_RECV },	/* 18 */
	{ "FAS_LINE_INQUIRY", FAS_ENABLE, FAS_LINE_INQUIRY },	/* 19 */
	{ "FAS_LPT_TEST", FAS_ENABLE, FAS_ENABLE },	/* 20 */
	{ "FAS_UNIT_POLL", FAS_UNIT_POLL, FAS_ENABLE },	/* 21 */
	{ "FAS_UNIT_UNLOCK", FAS_UNIT_UNLOCK, FAS_UNIT_UNLOCK },/* 22 */
	{ "FAS_UNIT_INQUIRY", FAS_UNIT_INQUIRY, FAS_UNIT_INQUIRY },/* 23 */
	{ "FAS_EXT_PARAMS", FAS_ENABLE, FAS_ENABLE },	/* 24 */
	{ "FAS_UNIT_RESET", FAS_ENABLE, FAS_ENABLE },	/* 25 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 26 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 27 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 28 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 29 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 30 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 31 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 32 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 33 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 34 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 35 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 36 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 37 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 38 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 39 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 40 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 41 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 42 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 43 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 44 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 45 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 46 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 47 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 48 */
	{ "Unknown", FAS_ENABLE, FAS_ENABLE },		/* 49 */
	{ "FAS_CONFIG_LINK", FAS_CONFIG_LINK, FAS_ENABLE },	/* 50 */
	{ "FAS_MUX_STATS", FAS_UNIT_STATS, FAS_RECV }		/* 51 */
};

/* adds a FAS command to the doubly linked list of queued cmds */
void add_cmd(sts_state * sts, char * data, int len) {
	cmd_list * cmd;

	if (!sts) {
		fprintf(stderr,"add_cmd: Got null sts\n");
		return;
	}
	if (!data && len!=0) {
		fprintf(stderr,"add_cmd: Got null data pointer\n");
		return;
	}
	if (!(cmd=malloc(sizeof(cmd_list)))) {
		perror("add_cmd pointer malloc");
		return;
	}
	if (!(cmd->data=malloc(len))) {
		perror("add_cmd data malloc");
		return;
	}
	memcpy(cmd->data, data, len);
	cmd->size=len;
	cmd->next=NULL;
	cmd->prev=sts->tail;
	if (sts->tail) sts->tail->next=cmd;
	sts->tail=cmd;
	if (!sts->head) sts->head=cmd;
}

/* find a properly size cmd in the list, starting with "cmd" */
cmd_list * find_cmd(sts_state * sts, cmd_list *cmd, int slotsize) {
	if (!sts) {
		fprintf(stderr,"find_cmd: Got null sts\n");
		return NULL;
	}
	if (!cmd) cmd=sts->head;
	for (;cmd;cmd=cmd->next)
		if (cmd->size<=slotsize) break;
	return cmd;
}

/* throws away a cmd in the list */
void delete_cmd(sts_state * sts, cmd_list * cmd) {
	if (!sts) {
		fprintf(stderr,"delete_cmd: Got null sts\n");
		return;
	}
	if (!cmd) {
		fprintf(stderr,"delete_cmd: Got null cmd\n");
		return;
	}
	if (cmd->prev) cmd->prev->next=cmd->next;
	if (cmd->next) cmd->next->prev=cmd->prev;
	if (sts->head == cmd) sts->head=cmd->next;
	if (sts->tail == cmd) sts->tail=cmd->prev;
	free(cmd->data);
	free(cmd);
}

void printchars(unsigned char * str, int len) {
	int i;

	for (i=0;i<len;i++) {
		if (isprint(str[i])) {
 		  sts_debug("\t\t%d: 0x%02X (%c)\n",i, str[i], str[i]);
		}
		else {
 		  sts_debug("\t\t%d: 0x%02X\n",i, str[i]);
		}
	}
}


/* FIXME: uuuuh... do I need to specify LUNs or does sg handle that? */
/* request vendor brand and model */
unsigned char * Inquiry (int fd) {
#define INQUIRY_CMD     0x12
#define INQUIRY_CMDLEN  6
#define INQUIRY_REPLY_LEN 96
  unsigned char cmd[MAX_PACKET];
  static unsigned char Inqbuffer[SG_LEN + INQUIRY_REPLY_LEN];
  unsigned char cmdblk[INQUIRY_CMDLEN] = { 
    INQUIRY_CMD,		/* command */
    0,				/* lun/reserved */
    0,				/* page code */
    0,				/* reserved */
    INQUIRY_REPLY_LEN,		/* allocation length */
    0				/* reserved/flag/link */
  };


  memcpy (cmd + SG_LEN, cmdblk, sizeof (cmdblk));

#ifdef SCSI_DEBUG
  sts_debug("scsi inquiry...\n");
#endif
  if (waitfor_scsi_cmd (fd, sizeof (cmdblk), 0, cmd,
		       sizeof (Inqbuffer) - SG_LEN, Inqbuffer)) {
      fprintf (stderr, "Inquiry failed\n");
      exit (2);
  }

  return Inqbuffer+SG_LEN;
}


/* test unit ready: 0 is okay? */
int TestForMedium (int fd) {
#define TESTUNITREADY_CMD 0
#define TESTUNITREADY_CMDLEN 6
#define ADD_SENSECODE 12
#define ADD_SC_QUALIFIER 13
#define NO_MEDIA_SC 0x3a
#define NO_MEDIA_SCQ 0x00
  unsigned char cmd[MAX_PACKET];
  /* request READY status */
  static unsigned char cmdblk[TESTUNITREADY_CMDLEN] = {
    TESTUNITREADY_CMD,		/* command */
    0,				/* lun/reserved */
    0,				/* reserved */
    0,				/* reserved */
    0,				/* reserved */
    0
  };				/* reserved */

  memcpy (cmd + SG_LEN, cmdblk, sizeof (cmdblk));

#ifdef SCSI_DEBUG
  sts_debug("scsi unitready...\n");
#endif
  if (waitfor_scsi_cmd (fd, sizeof (cmdblk), 0, cmd, 0, NULL)) {
      fprintf (stderr, "Test unit ready failed\n");
      exit (2);
  }

  return
    *(((struct sg_header *) cmd)->sense_buffer + ADD_SENSECODE) !=
    NO_MEDIA_SC ||
    *(((struct sg_header *) cmd)->sense_buffer + ADD_SC_QUALIFIER) !=
    NO_MEDIA_SCQ;
}

/* sts_write needs to act like the standard system write call */
int sts_write(sts_state * sts, unsigned char * fasblock, int len) {
#define SENDMSG_CMD	0x0A
#define SENDMSG_CMDLEN	6
  static unsigned char sendblk[SENDMSG_CMDLEN] = {
	SENDMSG_CMD,	/* Op Code */
	0,		/* Lun/reserved  FIXME? */
	0,0,0,          /* transfer length (needs to be set) */
	0		/* Control ?! */
  };
  /* FIXME: verify somewhere that MAX_PACKET is large enough
     for FAS_PKT_MAX + SG_LEN + SENDMSG_CMDLEN */
  static unsigned char cmd[MAX_PACKET];
  int result;
  unsigned char buffer[FAS_PKT_MAX + SG_LEN];

  if (!sts) {
	fprintf(stderr,"sts_write: got null sts!\n");
	errno=EBADF;
	return -1;
  }

  if (len > FAS_PKT_MAX) {
	fprintf(stderr,"send_fas_block: block too large: %d > %d\n",
		len, FAS_PKT_MAX);
	errno=ENOSPC;
	return -1;
  }

  /* read incoming FAS block from ELS and process */
  len=find_fas_block_size(sts,fasblock,len,
	cmd + SG_LEN + sizeof(sendblk),FAS_PKT_MAX,
	FAS_ELS);

  /* set the length bytes */
  sendblk[2]=((len & 0xff0000) >> 16);
  sendblk[3]=((len & 0xff00) >> 8);
  sendblk[4]=(len & 0xff);

  /* put the cmd block into the buffer also */
  memcpy (cmd + SG_LEN, sendblk, sizeof (sendblk));
  /* already did this during find_fas_block_size
  memcpy (cmd + SG_LEN + sizeof(sendblk), fasblock, len -1);
  */

#if 0
  /* FIXME: make the report-only var a define here */
  process_fas_block(sts,cmd+SG_LEN+sizeof(sendblk),len,1);
#endif

#ifdef SCSI_DEBUG
  sts_debug("sts_write ...\n");
  sts_debug("\tSG header\n");
  printchars(cmd,SG_LEN);
  sts_debug("\tSCSI cmd\n");
  printchars(cmd+SG_LEN,sizeof(sendblk));
  sts_debug("\tSCSI data\n");
  printchars(cmd+SG_LEN+sizeof(sendblk),len);
#endif
  result=waitfor_scsi_cmd(	
			sts->senddev->fd,
			SENDMSG_CMDLEN,
			len,
			cmd,
			0,
			buffer);

  if (result!=0) {
	fprintf(stderr,"sts_write: SCSI transport failure\n");
	errno=EIO;
	return -1;
  }

  return len;
}


/* send SCSI command to get info from STS */
int recv_fas_block(sts_state * sts) {
#define GETMSG_CMD	0x08
#define GETMSG_CMDLEN	6
  static unsigned char recvblk[GETMSG_CMDLEN] = {
	GETMSG_CMD,	/* Op Code */
	0,		/* Lun/reserved  FIXME? */
	0,8,0,          /* transfer len: 2048 */
	0		/* Control ?! */
  };
  static unsigned char cmd[MAX_PACKET];
  int result;

  if (!sts) {
	fprintf(stderr,"recv_fas_block: got null sts!\n");
	return 1;
  }

  memcpy(cmd + SG_LEN, recvblk, sizeof (recvblk));

#ifdef SCSI_DEBUG
  sts_debug("recv_fas_block ...\n");
#endif
  result=start_scsi_cmd(
			sts->recvdev->fd,
			GETMSG_CMDLEN,
			0,
			cmd,
			FAS_PKT_MAX);

  if (result!=0) {
	fprintf(stderr,"recv_fas_block: SCSI transport failure\n");
  }
  return result;
}

/* examine "block" for FAS commands, copying them into "fascmds" */
int find_fas_block_size(sts_state * sts, char * block, int maxsize,
  char * fascmds, int bufsize, int fromscsi) {
  int remaining;          /* how much left to read */
  unsigned char * ptr;    /* string pointer into block */
  int cmdat;                      /* debug: start of command */
  unsigned char * cmd;		/* current fas cmd */
  int len;                        /* how much data follows FAS_RECV */
  int size;
  int sawpktend=0;
  unsigned char opcode, port, status, follow;


  if (!fascmds || (bufsize < maxsize)) {
	fprintf(stderr,"find_fas_block: problem, incoming buffer too small\n");
	/* behave like read */
	errno=EFAULT;
	return -1;
  }

  if (!sts) {
	fprintf(stderr,"find_fas_block_size: got null sts!\n");
	errno=EINVAL;
	return -1;
  }
  if (!block && maxsize!=0) {
	fprintf(stderr,"find_fas_block_size: got null block!\n");
	errno=EINVAL;
	return -1;
  }

#if 0
	if (maxsize>0 && *block!=PKT_END)
#endif
	sts_debug("from %s (%d byte%s):\n",
		fromscsi ? "SCSI bus" : "ELS driver",maxsize,maxsize==1?"":"s");

        cmdat=0;
        ptr=block;
	cmd=fascmds;
        remaining=maxsize;
        size=0;
        for (;remaining;cmdat+=8) {
                opcode=*cmd=*ptr++;
                if (opcode == PKT_END) {
#if 0
			if (cmdat>0)
#endif
                        sts_debug("\tat %d: PKT_END\n",cmdat);
                        /* empty packet? */
                        if (cmdat == 0) {
                                sts->flushed=1;
                        }
			sawpktend=1;
			size++; /* got PKT_END... */
                        break; /* stop processing */
                }
                if (opcode == PKT_ERR) {
                        fprintf(stderr,"process_fas_block: buffer read "
                                       "overrun\n");
			/*
			errno=EFAULT;
                        return -1;
			*/
			/* be nicer: pass what we have */
			break;
                }
                if (remaining < FAS_CMD_LEN) {
                        fprintf(stderr,"process_fas_block: (opcode 0x%02X) "
				"buffer read underrun\n",opcode);
			/*
			errno=EFAULT;
                        return -1;
			*/
			/* be nicer: pass what we have */
			break;
                }
                memcpy(cmd+1,ptr,FAS_CMD_LEN - 1);
                ptr+=FAS_CMD_LEN-1;
                remaining-=FAS_CMD_LEN;
		size+=FAS_CMD_LEN;

		port=cmd[1];
		status=cmd[2];
		follow=cmd[3];

		if (opcode < FAS_CMD_MAX) {
		   /* handle any weird protocol translations */
		   if (fromscsi) {
#ifdef UNIT_POLL_HACK
			   /* trick ELS into hearing polls by reseting
			      status flag to 0 ? */
			   if (opcode == FAS_UNIT_POLL)
				status=cmd[2]=0;
#endif
		   }
		   else {
#if 0   /* doesn't help... */
			/* STS seems to think "0x81" is okay for DTR/RTS
			   but I don't think so... */
			if (opcode == FAS_SET_MODEM) {
				cmd[2]&=(~0x80);
				cmd[3]&=(~0x80);
			}
#endif
		   }
			
		   sts_debug("\tat %d: %s (%04X %04X %04X %04X)\n",
			cmdat, fas_opcodes[opcode].t, 
			(cmd[0] << 8)|(cmd[1]),
			(cmd[2] << 8)|(cmd[3]),
			(cmd[4] << 8)|(cmd[5]),
			(cmd[6] << 8)|(cmd[7]));

                   if ((!fromscsi && fas_opcodes[opcode].req_family == FAS_SEND) ||
		       (fromscsi && fas_opcodes[opcode].rsp_family == FAS_RECV)) {
	                len=GRAB_16BIT(cmd,fromscsi ? 4 : 2);

                        if (remaining < len) {
                                   fprintf(stderr,"process_fas_block: %s "
                                        "data underrun -- expected %d got %d "
                                        "or less\n", fas_opcodes[opcode].t,
                                        len,remaining);
				   errno=EFAULT;
                                   return -1;
                         }

                         if (len>0) {
					size+=len;
                                        memcpy(cmd+FAS_CMD_LEN,ptr,len);
                                        ptr+=len;
                                        remaining-=len;

                                        sts_debug("\tat %d: %s data (%d "
                                           "byte%s)\n",
                                           cmdat+FAS_CMD_LEN,
					   fas_opcodes[opcode].t,
                                           len, len == 1 ? "" : "s");
                                        cmdat+=len;
                                        printchars(cmd+FAS_CMD_LEN,len);
			                cmd+=len;
                          }
			  else {
                                        sts_debug("\tat %d: %s has no data\n",
                                           cmdat+FAS_CMD_LEN,
					   fas_opcodes[opcode].t);
			  }
		   }
		}
		else {
			sts_debug("\tat %d: 0x%02X\n", cmdat, opcode);
		}

		cmd+=FAS_CMD_LEN;
	}

  if (fromscsi) {
	/* if it's from the SCSI bus, we don't want the PKT_END */
	if (sawpktend) size--;
  }
  else {
	/* tack on trailing PKT_END if not at packet max */
	if (!sawpktend && size>0 && size<FAS_PKT_MAX) {
		sts_debug("\tat %d: PKT_END (inserted)\n",cmdat);
		*cmd=PKT_END;
		size++;
	}
  }

  if (size > 0) 
  	sts_debug("sending %d byte%s to %s\n",size, size == 1 ? "" : "s",
		!fromscsi ? "SCSI bus" : "ELS driver");

#ifdef STS_DEBUG
  printchars(fascmds,size);
#endif

  return size;
}

/* act like system read call */
int sts_read(sts_state * sts, char * data_buf, int n) {
  int result;
  unsigned char buffer[FAS_PKT_MAX + SG_LEN+1];

  if (!sts) {
	fprintf(stderr,"sts_read: got null sts!\n");
	errno=EBADF;
	return -1;
  }
  if (!data_buf) {
	fprintf(stderr,"sts_read: got null data_buf!\n");
	errno=EFAULT;
	return -1;
  }
  if (n<FAS_PKT_MAX) {
	fprintf(stderr,"sts_read: input buffer too small!\n");
	errno=ENOMEM;
  }

  /* fill overflow data with PKT_ERRs for safety */
  memset(buffer+SG_LEN,PKT_ERR,FAS_PKT_MAX+1);

#ifdef SCSI_DEBUG
  sts_debug("sts_read ...\n");
#endif
  result=finish_scsi_cmd(
			sts->recvdev->fd,
			FAS_PKT_MAX,
			buffer);

  if (result!=0) {
	fprintf(stderr,"sts_read: SCSI transport failure\n");
	errno=EIO;
	return -1;
  }
  /* immediately re-post the recv request */
  recv_fas_block(sts);

  return find_fas_block_size(sts,
	buffer+SG_LEN,FAS_PKT_MAX,
	data_buf,n,
	FAS_SCSI);
}

sts_state * sts_shutdown(sts_state * sts) {
  int result;
  unsigned char buffer[FAS_PKT_MAX + SG_LEN];

  if (!sts) {
	fprintf(stderr,"sts_read: got null sts!\n");
  }
  else {

#ifdef SCSI_DEBUG
	sts_debug("sts_shutdown ...\n");
#endif
	result=finish_scsi_cmd(
			sts->recvdev->fd,
			FAS_PKT_MAX,
			buffer);

	if (result!=0) {
		fprintf(stderr,"sts_shutdown: SCSI transport failure\n");
	}

	scsi_close(sts->senddev);
	scsi_close(sts->recvdev);
	free(sts);
  }

  return NULL;
}

